# Copyright (c) 2024, InfinityQ Technology, Inc.

"""
Errors specific to the TitanQ SDK.
"""

class TitanqError(Exception):
    """Base TitanQ error"""

class InvalidUrl(TitanqError):
    """API URL is not valid or accessible"""

class UnexpectedServerResponseError(TitanqError):
    """Response from the TitanQ server is not as expected"""

class MissingTitanqApiKey(TitanqError):
    """TitanQ Api key is missing"""

class MissingVariableError(TitanqError):
    """Variable has not already been registered"""

class VariableAlreadyExist(TitanqError):
    """Variable with the same name already exist"""

class MissingObjectiveError(TitanqError):
    """Objective has not already been registered"""

class MaximumConstraintLimitError(TitanqError):
    """The number of constraints is bigger than the number of variables"""

class ConstraintSizeError(TitanqError):
    """Unexpected number of constraints"""

class ConstraintAlreadySetError(TitanqError):
    """A constraint has already been set"""

class ObjectiveAlreadySetError(TitanqError):
    """An objective has already been set"""

class OptimizeError(TitanqError):
    """Error occur during optimization"""

class ServerError(TitanqError):
    """Unexpected condition prevented the TitanQ server to fulfill the request"""

class ConnectionError(TitanqError):
    """Error due to a connection issue with an external resource"""

class UnsolvableRequestError(TitanqError):
    """TitanQ cannot solve this combination of parameters"""

    def __init__(self, message="TitanQ cannot solve this combination of parameters", *args, **kwargs):
        super().__init__(message, *args, **kwargs)

class TautologicalExpressionError(TitanqError):
    """
    Exception raised when an expression is tautological (always true).

    This exception indicates that the provided expression is redundant
    and does not add meaningful constraints or information.
    """
    def __init__(self, message="The provided expression is tautological and always evaluates to True, regardless of the variable values.", *args, **kwargs):
        super().__init__(message, *args, **kwargs)

class ContradictoryExpressionError(TitanqError):
    """
    Exception raised when an expression is contradictory (always false).

    This exception indicates that the provided expression is invalid
    as it represents an impossible condition.
    """
    def __init__(self, message="The provided expression is contradictory and always evaluates to False, regardless of the variable values.", *args, **kwargs):
        super().__init__(message, *args, **kwargs)

class MpsParsingError(TitanqError):
    """Base class for any error related to the MPS files parsing module"""

class MpsConfiguredModelError(MpsParsingError):
    """Passed model is already configured"""

class MpsMissingValueError(MpsParsingError):
    """A required value is missing"""

class MpsMissingSectionError(MpsParsingError):
    """A required section is missing"""

class MpsMalformedFileError(MpsParsingError):
    """The file is malformed"""

class MpsUnexpectedValueError(MpsParsingError):
    """Found an unexpected value"""

class MpsUnsupportedError(MpsParsingError):
    """Found an unsupported value"""

class ClientError(TitanqError):
    """Base client-side http error"""

class NotEnoughCreditsError(ClientError):
    """Not enough credits left"""

    def __init__(self, message="Not enough credits left", *args, **kwargs):
        super().__init__(message, *args, **kwargs)

class BadRequest(ClientError):
    """The request sent to TitanQ is not valid"""


# Copyright (c) 2024, InfinityQ Technology, Inc.
import datetime
from io import BytesIO
from itertools import chain
import boto3
import boto3.s3
import boto3.s3.transfer
import botocore
import logging
import time

from .._client.model import AwsStorage, S3Input, S3Output
from .storage_client import Ftype, StorageClient

log = logging.getLogger("TitanQ")

_TRANSFER_CONFIG = boto3.s3.transfer.TransferConfig(
    multipart_chunksize=5 * 1024 * 1025, # 5 MB
    multipart_threshold=5 * 1024 * 1024, # 5 MB
)

class S3Storage(StorageClient):
    """Storage client using S3 bucket from AWS"""
    def __init__(
        self,
        access_key: str,
        secret_key: str,
        bucket_name: str,
    ) -> None:
        """
        Initiate the S3 bucket client for handling temporary files.

        Parameters
        ----------
        access_key
            Used to upload and download files from an AWS S3 bucket.
        secret_key
            Used to upload and download files from an AWS S3 bucket.
        bucket_name
            Name of the AWS S3 bucket used to store temporarily data that the TitanQ optimizer will read.

        Raises
        ------
        botocore.exceptions.ParamValidationError
            If any AWS argument is missing this will raise an exception.

        Examples
        --------
        >>> storage_client = S3Storage(
        >>>     access_key="{insert aws bucket access key here}",
        >>>     secret_key="{insert aws bucket secret key here}",
        >>>     bucket_name="{insert bucket name here}"
        >>> )
        """
        self._s3 = boto3.client('s3', aws_access_key_id=access_key, aws_secret_access_key=secret_key)
        self._access_key_id = access_key
        self._secret_access_key = secret_key
        self._bucket_name = bucket_name

        timestamp = datetime.datetime.now().isoformat()
        self._remote_folder = f"titanq_sdk/{timestamp}"

        # keep track of which file were uploaded
        self._file_uploaded = set()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self._delete_remote_object()

    def upload(self, file_type: Ftype, data: bytes):
        # check if the file type is one that can be uploaded. if so, do so
        if file_type not in (Ftype.WEIGHTS,
                         Ftype.BIAS,
                         Ftype.VARIABLE_BOUNDS,
                         Ftype.CONSTRAINT_WEIGHTS,
                         Ftype.CONSTRAINT_BOUNDS,
                         Ftype.QUAD_CONSTRAINT_WEIGHTS,
                         Ftype.QUAD_CONSTRAINT_BOUNDS,
                         Ftype.QUAD_CONSTRAINT_LINEAR_WEIGHTS,):
            raise NotImplementedError(f"File type {file_type} not supported for upload")

        filename = self._get_full_filename(file_type.value)

        log.debug(f"Uploading object on AWS s3: {filename}")
        temp_file = BytesIO(data)
        self._s3.upload_fileobj(temp_file, Bucket=self._bucket_name, Key=filename, Config=_TRANSFER_CONFIG)
        self._file_uploaded.add(filename)


    def input(self) -> S3Input:
        weights = self._get_full_filename(Ftype.WEIGHTS.value)
        variable_bounds = self._get_full_filename(Ftype.VARIABLE_BOUNDS.value)
        constraint_weights = self._get_full_filename(Ftype.CONSTRAINT_WEIGHTS.value)
        constraint_bounds = self._get_full_filename(Ftype.CONSTRAINT_BOUNDS.value)
        quad_constraint_weights = self._get_full_filename(Ftype.QUAD_CONSTRAINT_WEIGHTS.value)
        quad_constraint_bounds = self._get_full_filename(Ftype.QUAD_CONSTRAINT_BOUNDS.value)
        quad_constraint_linear_weights = self._get_full_filename(Ftype.QUAD_CONSTRAINT_LINEAR_WEIGHTS.value)

        return S3Input(
            s3=self._get_api_model_location(),
            weights_file_name=weights if weights in self._file_uploaded else None,
            bias_file_name=self._get_full_filename(Ftype.BIAS.value),
            variable_bounds_file_name=variable_bounds if variable_bounds in self._file_uploaded else None,
            constraint_weights_file_name=constraint_weights if constraint_weights in self._file_uploaded else None,
            constraint_bounds_file_name=constraint_bounds if constraint_bounds in self._file_uploaded else None,
            quad_constraint_weights_file_name=quad_constraint_weights if quad_constraint_weights in self._file_uploaded else None,
            quad_constraint_bounds_file_name=quad_constraint_bounds if quad_constraint_bounds in self._file_uploaded else None,
            quad_constraint_linear_weights_file_name=quad_constraint_linear_weights if quad_constraint_linear_weights in self._file_uploaded else None,
            manifest=None
        )

    def output(self) -> S3Output:
        return S3Output(
            result_archive_file_name=self._get_full_filename(Ftype.RESULT.value),
            s3=self._get_api_model_location())

    def _get_api_model_location(self) -> AwsStorage:
        """
        :return: An AwsStorage object that can be used in the api_model using this S3 credentials
        """
        return AwsStorage(
            bucket_name=self._bucket_name,
            access_key_id=self._access_key_id,
            secret_access_key=self._secret_access_key,
        )

    def wait_for_result_to_be_uploaded_and_download(self) -> bytes:
        result_file = self._get_full_filename(Ftype.RESULT.value)
        self._wait_for_file_to_be_uploaded(result_file)
        return self._download_file(result_file)

    def _wait_for_file_to_be_uploaded(self, filename: str):
        """
        Wait until a file exist in a bucket. It also verifies if the
        file is bigger than 0 bytes, this will ensure not downloading
        the empty archive file uploaded to test credentials

        :param filename: The full path of the file that is uploaded
        """
        log.debug(f"Waiting until object get upload on AWS s3: {filename}")
        while True:
            try:
                # check if file exist in the s3 bucket
                response = self._s3.head_object(
                    Bucket=self._bucket_name,
                    Key=filename,
                )
                # check if file content_length > 0
                if response['ContentLength'] > 0:
                    break
            except botocore.exceptions.ClientError as ex:
                # if the error we got is not 404, This is an unexpected error. Raise it
                if int(ex.response['Error']['Code']) != 404:
                    raise

            time.sleep(0.25) # wait 0.25 sec before trying again

    def _download_file(self, filename) -> bytes:
        """
        Download file from remote s3 bucket

        :param filename: The full path of the file to be uploaded

        :return: content of the file
        """
        log.debug(f"Downloading object from AWS s3: {filename}")
        object = self._s3.get_object(Bucket=self._bucket_name, Key=filename)
        return object['Body'].read()

    def _delete_remote_object(self):
        """
        Delete remote object on AWS s3

        :param key: object name to be deleted on the remote s3 bucket
        """
        for file_name in chain(self._file_uploaded, [self._get_full_filename(Ftype.RESULT.value)]):
            log.debug(f"Deleting object on AWS s3: {file_name}")
            self._s3.delete_object(Bucket=self._bucket_name, Key=file_name)

    def _get_full_filename(self, filename: str) -> str:
        return f"{self._remote_folder}/{filename}"



# Copyright (c) 2024, InfinityQ Technology, Inc.

from abc import ABC, abstractmethod
import enum
from typing import Tuple

import numpy.typing as np_type
import numpy as np
import symengine as se
import symengine.lib.symengine_wrapper as se_type

class Vtype(str, enum.Enum):
    """
    All variable types currently supported by the solver.

    ℹ️ **NOTE:** Bipolar variables are not directly supported,
    but :class:`tools.BipolarToBinary` can be used as an alternative.
    """

    BINARY = 'binary'
    INTEGER = 'integer'
    CONTINUOUS = 'continuous'

    def __str__(self) -> str:
        return str(self.value)

    def _api_str(self) -> str:
        """
        Convert the `Vtype` enum to its corresponding API string representation.

        This method returns a shorthand character used by the API to identify 
        the variable type. The mapping is as follows:

        - 'b' for BINARY
        - 'c' for CONTINUOUS
        - 'i' for INTEGER

        Returns
        -------
        A string representing the variable type in the API.
        """
        if self == Vtype.BINARY:
            return 'b'
        elif self == Vtype.CONTINUOUS:
            return 'c'
        else: # Vtype.Integer
            return 'i'


class Variable(ABC, se_type.Symbol):
    def __init__(self, parent_name: str, parent_index: int, problem_index: int,  *args, **kwargs):
        name = f"{parent_name}[{parent_index}]"
        super().__init__(name, *args, **kwargs)

        self._problem_index = problem_index
        self._parent_name = parent_name

    @abstractmethod
    def vtype(self) -> Vtype:
        """
        Returns
        -------
        Type of the variable.
        """

    @abstractmethod
    def variable_bounds(self) -> np_type.NDArray[np.float32]:
        """
        Returns
        -------
        The variable bounds associated to this variable
        """

    def problem_index(self) -> int:
        """
        Returns
        -------
        The index of this variable for the whole problem
        """
        return self._problem_index

    def parent_name(self) -> str:
        """
        Returns
        -------
        The name of the parent variable vector
        """
        return self._parent_name


class BinaryVariable(Variable):
    def vtype(self) -> Vtype:
        return Vtype.BINARY

    def variable_bounds(self) -> np_type.NDArray[np.float32]:
        return np.array([0,1], dtype=np.float32)


class IntegerVariable(Variable):
    def __init__(self, parent: str, index: int, problem_index: int, bounds: Tuple[int, int]) -> None:
        super().__init__(parent, index, problem_index)
        self._bounds = bounds

    def vtype(self) -> Vtype:
        return Vtype.INTEGER

    def variable_bounds(self) -> np_type.NDArray[np.float32]:
        return self._bounds

class ContinuousVariable(Variable):
    def __init__(self, parent: str, index: int, problem_index: int, bounds: Tuple[int, int]) -> None:
        super().__init__(parent, index, problem_index)
        self._bounds = bounds

    def vtype(self) -> Vtype:
        return Vtype.CONTINUOUS

    def variable_bounds(self) -> np_type.NDArray[np.float32]:
        return self._bounds


# Copyright (c) 2024, InfinityQ Technology, Inc.
import enum
from typing import Any, List


class RowsType(str, enum.Enum):
    """
    All supported Rows type for .mps files in this parser
    """
    FREE_ROW = 'N'
    GREATER_OR_EQUAL = 'G'
    LOWER_OR_EQUAL = 'L'
    EQUALITY = 'E'


class ColumnsType(str, enum.Enum):
    """
    ALl supported Columns type for .mps files in this parser
    """
    CONTINUOUS = 'c'
    INTEGER = 'i'


class BoundsType(str, enum.Enum):
    """
    All supported Bounds type for .mps files in this parser
    """
    LOWER_BOUND = 'LO'
    LOWER_BOUND_INT = 'LI'
    UPPER_BOUND = 'UP'
    UPPER_BOUND_INT = 'UI'
    FIXED_VALUE = 'FX'
    FREE_VARIABLE = 'FR'
    MINUS_INFINITY = 'MI'
    PLUS_INFINITY = 'PL'
    BINARY_VARIABLE = 'BV'
    SEMI_CONTINUOUS = 'SC'

    @classmethod
    def needs_value(cls) -> List['BoundsType']:
        """returns a list of the bounds type that needs a value assigned to them"""
        return [cls.LOWER_BOUND, cls.LOWER_BOUND_INT, cls.UPPER_BOUND, cls.UPPER_BOUND_INT, cls.FIXED_VALUE]


class SectionType(str, enum.Enum):
    """
    All supported Sections type for .mps files in this parser
    """
    NAME = "NAME"
    ROWS = "ROWS"
    COLUMNS = "COLUMNS"
    RHS = "RHS"
    RANGES = "RANGES"
    BOUNDS = "BOUNDS"
    QUADOBJ = "QUADOBJ"
    QMATRIX = "QMATRIX"
    ENDATA = "ENDATA"


class UniqueList:
    def __init__(
        self,
        error_msg: str
    ) -> None:
        """
        Unique list is an iterable containing a list that can only have unique values.
        """
        self._list = []
        self._set = set()
        self._error_msg = error_msg

    def append(self, value: Any) -> None:
        if value in self._set:
            raise Exception(
                f"{self._error_msg}: Tried to append an already existing value in a {self.__class__.__name__} '{value}'")
        self._list.append(value)
        self._set.add(value)

    def __iter__(self):
        return iter(self._list)


# Copyright (c) 2024, InfinityQ Technology, Inc.
import logging
import requests
import time
import urllib3

from .._client.model import UrlInput, UrlOutput
from .storage_client import Ftype, StorageClient
from .._client.client import Client
from ..errors import ConnectionError

log = logging.getLogger("TitanQ")


class ManagedStorage(StorageClient):
    def __init__(self, titanq_client: Client):
        """
        Initiate the managed storage client for handling the temporary files.

        :titanq_client: titanq_client to be used to fetch temporary URL's
        """
        self._titanq_client = titanq_client
        self._urls = None
        self._initialize_upload_status()
        self._session = None

    def __enter__(self):
        self._initialize_upload_status()

        self._session = requests.Session()
        retries = urllib3.Retry(
                    total=3,
                    backoff_factor=0.5,
                    status_forcelist=[502, 503, 504, 495],
                )
        self._session.mount('https://', requests.adapters.HTTPAdapter(max_retries=retries))

        self._urls = self._titanq_client.temp_storage()

        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self._delete_remote_object()
        if self._session: self._session.close()

    def upload(self, file_type: Ftype, data: bytes):
        url, has_a_flag_to_set = {
            Ftype.WEIGHTS: (self._urls.input.weights_file.upload, True),
            Ftype.BIAS: (self._urls.input.bias_file.upload, False),
            Ftype.VARIABLE_BOUNDS: (self._urls.input.variable_bounds_file.upload, True),
            Ftype.CONSTRAINT_BOUNDS: (self._urls.input.constraint_bounds_file.upload, True),
            Ftype.CONSTRAINT_WEIGHTS: (self._urls.input.constraint_weights_file.upload, True),
            Ftype.QUAD_CONSTRAINT_WEIGHTS: (self._urls.input.quad_constraint_weights_file.upload, True),
            Ftype.QUAD_CONSTRAINT_BOUNDS: (self._urls.input.quad_constraint_bounds_file.upload, True),
            Ftype.QUAD_CONSTRAINT_LINEAR_WEIGHTS: (self._urls.input.quad_constraint_linear_weights_file.upload, True),

        }.get(file_type, (None, None))

        if url:
            log.debug(f"Uploading file ({file_type.value}) to temporary storage")
            self._session.put(url, data=data)
            if has_a_flag_to_set and file_type in self._file_upload_status:
                self._file_upload_status[file_type] = True
        else:
            raise NotImplementedError(f"File type {file_type} not supported for upload")

    def input(self) -> UrlInput:
        return UrlInput(
            weights_file_name=self._urls.input.weights_file.download if self._file_upload_status[Ftype.WEIGHTS] else None,
            bias_file_name=self._urls.input.bias_file.download,
            variable_bounds_file_name=self._urls.input.variable_bounds_file.download if self._file_upload_status[Ftype.VARIABLE_BOUNDS] else None,
            constraint_weights_file_name=self._urls.input.constraint_weights_file.download if self._file_upload_status[Ftype.CONSTRAINT_WEIGHTS] else None,
            constraint_bounds_file_name=self._urls.input.constraint_bounds_file.download if self._file_upload_status[Ftype.CONSTRAINT_BOUNDS] else None,
            quad_constraint_weights_file_name=self._urls.input.quad_constraint_weights_file.download if self._file_upload_status[Ftype.QUAD_CONSTRAINT_WEIGHTS] else None,
            quad_constraint_bounds_file_name=self._urls.input.quad_constraint_bounds_file.download if self._file_upload_status[Ftype.QUAD_CONSTRAINT_BOUNDS] else None,
            quad_constraint_linear_weights_file_name=self._urls.input.quad_constraint_linear_weights_file.download if self._file_upload_status[Ftype.QUAD_CONSTRAINT_LINEAR_WEIGHTS] else None,
            manifest=None
        )

    def output(self) -> UrlOutput:
        return UrlOutput(result_archive_file_name=self._urls.output.result_archive_file.upload)

    def wait_for_result_to_be_uploaded_and_download(self) -> bytes:
        """
        Wait until the content of the file in the temporary storage is bigger
        than 0 bytes. Meaning it will wait until the file is uploaded

        :param url: Url to download the file.
        """
        retries = 0
        while retries < 5:
            try:
                response = self._session.get(self._urls.output.result_archive_file.download)
                response.raise_for_status()
                if len(response.content) > 0:
                    return response.content

            except (urllib3.exceptions.ProtocolError, requests.exceptions.ConnectionError, ConnectionResetError) as e:
                retries = retries + 1
                log.warning(f"Caught error {e} [retries: {retries}]")
                time.sleep(1.0)

            time.sleep(0.25)

        raise ConnectionError("Unexpected error with InfinityQ internal storage, please contact InfinityQ support for more information")

    def _delete_remote_object(self) -> None:
        log.debug("Temporary storage option does not delete any file at the moment")

    def _initialize_upload_status(self):
        """Initializes or resets the upload status dictionary."""
        self._file_upload_status = {
            Ftype.WEIGHTS: False,
            Ftype.VARIABLE_BOUNDS: False,
            Ftype.CONSTRAINT_WEIGHTS: False,
            Ftype.CONSTRAINT_BOUNDS: False,
            Ftype.QUAD_CONSTRAINT_WEIGHTS: False,
            Ftype.QUAD_CONSTRAINT_BOUNDS: False,
            Ftype.QUAD_CONSTRAINT_LINEAR_WEIGHTS: False,
        }



# Copyright (c) 2024, InfinityQ Technology, Inc.

"""
This module offers methods to define a :class:`Model` from a MPS file.
"""

from collections import defaultdict
import os
from pathlib import Path
from typing import Dict, List, Optional, Tuple, Union

from deprecated.sphinx import deprecated
import numpy as np

from titanq import Model, Vtype

from .bounds import ConstraintBounds, VariableBounds
from titanq.errors import MpsConfiguredModelError, MpsUnexpectedValueError, MpsUnsupportedError
from ..mps_parser import (
    BoundsType,
    ColumnsType,
    MPSBounds,
    MPSColumns,
    MPSObject,
    MPSQuadobj,
    MPSParseOptions,
    MPSRhs,
    parse_from_mps,
    RowsType
)
from .utils import SetOnceOrDefault, get_variables_and_types


def from_mps(
    path: Union[str, os.PathLike],
    model: Model,
    *,
    skip_empty_lines: bool = False,
) -> None:
    """
    Configure a model from an .mps file. It currently supports these following sections.

    - NAME: The name of the problem.
    - ROWS: The definition of constraints.
    - COLUMNS: The coefficients for the variables.
    - RHS: The right-hand side values for the constraints.
    - BOUNDS: The bounds on the variables.
    - QUADOBJ or QMATRIX: The quadratic objective matrix
    - ENDATA: Marks the end of the data.

    Integer variables in .mps files are supported both by the markers in the COLUMNS section 
    and from the types in the BOUNDS section.

    Parameters
    ----------
    path
        The path to the .mps file
    model
        The instance of the model to configure. Must not be already
        configured.
    options
        Additional options applied when parsing the .mps file

    Raises
    ------
    OSError, FileNotFoundError
        If the specified MPS file path cannot be open.
    :class:`titanq.errors.MpsConfiguredModelError`
        If the provided model already contains data (e.g. variables).
    :class:`titanq.errors.MpsParsingError`
        If the provided MPS file cannot be parsed.

    Example
    -------
    >>> from titanq.tools import from_mps
    >>> from titanq import Model
    >>> model = Model()
    >>> from_mps("path/to/file.mps", model)
    >>> model.optimize()
    """
    # validate the model is empty
    if len(model._variables) > 0:
        raise MpsConfiguredModelError(
            "Found variables in the model for the .mps file parser."
            "Did you pass a pre-configured model? The model must be empty."
        )

    mps_object = parse_from_mps(path, MPSParseOptions(skip_empty_lines=skip_empty_lines))

    # build a unique list of variables from the COLUMNS with their corresponding types
    # example [('COL01', ColumnsType.CONTINUOUS), ('COL02', ColumnsType.CONTINUOUS)]
    variables = get_variables_and_types(mps_object.columns)

    # build a dictionary mapping each variable name to their indices which allows
    # an average of O(1) instead of O(n) when trying to find the index of a variable identifier
    # example { 'COL1': 0, 'COL2': 1 }
    variables_map = { item[0]: index for index, item in enumerate(variables) }

    # create variables with their name, type and bounds
    _create_variables(model, variables, mps_object.bounds)

    # weights and bias
    weights = _get_weights(variables_map, mps_object.quadobj)
    bias = _get_bias(variables, mps_object.columns, mps_object.free_row.identifier())
    model.set_objective_matrices(weights, bias)

    # constraints
    _create_constraints(model, variables_map, mps_object)


@deprecated(version="0.26.0", reason="Use :func:`from_mps` instead.")
def configure_model_from_mps_file(model: Model, file_path: Path) -> None:
    """
    Deprecated. Alternative form of :func:`from_mps`.

    Parameters
    ----------
    model
        The instance of the model to configure.
    file_path
        The path to the MPS file.
    """
    from_mps(file_path, model)


def _get_weights(variables_map: Dict, quadobj: List[MPSQuadobj]) -> Optional[np.ndarray]:
    if len(quadobj) == 0:
        return None

    variables_length = len(variables_map)
    weights = np.zeros((variables_length, variables_length), dtype=np.float32)

    for row in quadobj:
        weights[variables_map[row.row_identifier()]][variables_map[row.column_identifier()]] = row.value()

    return weights


def _get_bias(variables: List[str], columns: List[MPSColumns], objective_identifier: str) -> np.ndarray:
    """obtain bias values from the COLUMNS section, the objective identifier is the bias"""
    variables_dict = defaultdict(lambda:
        SetOnceOrDefault(0, "Variable with more than one objective value", allow_same_value=True)
    )

    # read bias values from the columns
    for column in columns:
        if column.row_identifier() == objective_identifier:
            variables_dict[column.identifier()].set(column.coeff())

    bias = []
    for name, _ in variables:
        bias.append(variables_dict[name].get())

    return np.array(bias, dtype=np.float32)


def _create_variables(model: Model, variables: List[Tuple[str, ColumnsType]], bounds: List[MPSBounds]) -> None:
    """obtain the variables bounds from the BOUNDS section"""
    var_bounds_dict = defaultdict(lambda: VariableBounds(0.0, np.nan, Vtype.CONTINUOUS))
    for bound in bounds:
        type = bound.type()
        value = bound.value()
        var_bound = var_bounds_dict[bound.column_identifier()]

        # in the following statements, we want to override vtype for each entry
        # to avoid types mixins. Bounds object will not tolerate it
        if type == BoundsType.LOWER_BOUND:
            var_bound.set(lower=value, vtype=Vtype.CONTINUOUS)
        elif type == BoundsType.LOWER_BOUND_INT:
            var_bound.set(lower=value, vtype=Vtype.INTEGER)
        elif type == BoundsType.UPPER_BOUND:
            var_bound.set(upper=value, vtype=Vtype.CONTINUOUS)
        elif type == BoundsType.UPPER_BOUND_INT:
            var_bound.set(upper=value, vtype=Vtype.INTEGER)
        elif type == BoundsType.FIXED_VALUE: # upper and lower bound the same
            var_bound.set(lower=value, upper=value, vtype=Vtype.CONTINUOUS)
        elif type == BoundsType.FREE_VARIABLE: # lower bound -∞ and upper bound +∞
            var_bound.set(lower=-np.nan, upper=np.nan, vtype=Vtype.CONTINUOUS)
        elif type == BoundsType.MINUS_INFINITY: # lower bound = -∞
            var_bound.set(lower=-np.nan, vtype=Vtype.CONTINUOUS)
        elif type == BoundsType.PLUS_INFINITY: # upper bound = +∞
            var_bound.set(upper=np.nan, vtype=Vtype.CONTINUOUS)
        elif type == BoundsType.BINARY_VARIABLE:
            var_bound.set(vtype=Vtype.BINARY)
        elif type == BoundsType.SEMI_CONTINUOUS:
            raise MpsUnsupportedError(f"Type of bound '{BoundsType.SEMI_CONTINUOUS}' is not supported by TitanQ")
        else:
            # this should never happen, as it is handled by the parser
            raise Exception(f"Unknown type of bound '{type}'")

    # set variables
    for name, type in variables:
        variable_bound = var_bounds_dict[name]

        # unless the bounds are defined as BINARY, we override the variable type and the bounds if not set because
        # the variable type is defined in the columns section with a marker, this is the result of the integer extensions.
        variable_is_integer = not variable_bound.is_binary() and type == ColumnsType.INTEGER

        variable_bound.into_model_variable(model, name, variable_is_integer)


def _get_constraint_weights(
    variables_map: Dict[str, int],
    columns: List[MPSColumns],
    objective_identifier: str,
    constraints_name: List[str],
) -> None:
    """obtain the constraints weights from the COLUMN section"""
    variables_length = len(variables_map)
    constraint_weights = np.zeros((variables_length, variables_length), dtype=np.float32)
    for column in columns:
        # ignore objective column
        if column.row_identifier() == objective_identifier:
            continue

        column_index = variables_map[column.identifier()]
        row_index = constraints_name.index(column.row_identifier())
        constraint_weights[row_index][column_index] = column.coeff()

    return constraint_weights


def _get_constraint_bounds(rhs: List[MPSRhs], constraints_dict: dict, constraints_name: List[str]):
    """obtain the constraints bounds from the RHS and RANGES section"""
    constraints_bounds_dict = defaultdict(lambda: ConstraintBounds(0.0, 0.0))
    for r in rhs:
        constraint = constraints_dict[r.row_identifier()]
        sense = constraint["sense"]
        range = constraint["range"] if constraint["range"] else np.nan
        constraint_bound = constraints_bounds_dict[r.row_identifier()]

        if sense == RowsType.GREATER_OR_EQUAL: # [Lower: rhs, Upper: rhs + |range|]
            constraint_bound.set(lower=r.coeff(), upper=r.coeff() + abs(range))
        elif sense == RowsType.LOWER_OR_EQUAL: # [Lower: rhs - |range|, Upper: rhs]
            constraint_bound.set(lower=r.coeff() - abs(range), upper=r.coeff())
        elif sense == RowsType.EQUALITY:
            constraint_bound.set(lower=r.coeff(), upper=r.coeff())
        elif sense == RowsType.FREE_ROW:
            # this should never happen, as it is handled by the parser
            raise MpsUnexpectedValueError("Found a free row while trying to create constraint bounds")
        else:
            # this should never happen, as it is handled by the parser
            raise Exception(f"Unknown type of sense '{sense}'")

    constraint_bounds = []
    for constraint_name in constraints_name:
        lower, upper = constraints_bounds_dict[constraint_name].get()
        constraint_bounds.append([lower, upper])

    return constraint_bounds


def _create_constraints(model: Model, variables_map: Dict[str, int], mps: MPSObject) -> None:
    """Append the constraint weights and the constrain bounds to the model from the MPS object."""
    # pre process to iterate less. Find each of the constraints sense and range.
    # this will enable iterating over ROWS and RANGE only once
    constraints_dict = {}
    for row in mps.rows:
        constraints_dict[row.identifier()] = { "sense": row.sense(), "range": None }
    for range in mps.ranges:
        constraints_dict[range.row_identifier()]["range"] = range.coeff()

    # keys as a list
    constraints_name = list(constraints_dict.keys())

    # constraint weights
    constraint_weights = _get_constraint_weights(variables_map, mps.columns, mps.free_row.identifier(), constraints_name)

    # constraint bounds
    constraint_bounds = _get_constraint_bounds(mps.rhs, constraints_dict, constraints_name)

    # add constraints to the model
    for index, constraint in enumerate(constraints_dict):
        sense = constraints_dict[constraint]["sense"]
        if sense == RowsType.GREATER_OR_EQUAL or sense == RowsType.LOWER_OR_EQUAL:
            model.add_inequality_constraint(
                constraint_mask=np.array(constraint_weights[index], dtype=np.float32),
                constraint_bounds=np.array(constraint_bounds[index], dtype=np.float32)
            )
        elif sense == RowsType.EQUALITY:
            model.add_equality_constraint(
                constraint_mask=np.array(constraint_weights[index], dtype=np.float32),
                limit=np.array(constraint_bounds[index][0], dtype=np.float32) # does not matter here upper or lower
            )



# Copyright (c) 2024, InfinityQ Technology, Inc.

"""
TitanQ
======

The TitanQ SDK for Python. This package will let you use InfinityQ's solver named TitanQ.

Documentation
-------------

Documentation is available as docstrings.
See also: https://sdk.titanq.infinityq.io

License
-----------------------------

Apache Software License (Apache 2.0)

"""

# These symbols must be exposed by this lib
from ._model import fast_sum, fastSum, Model, OptimizeResponse, Precision, Target, Vtype
# S3Storage must be exposed to the end user if they wish to use s3 buckets
from ._storage import S3Storage

# logger config
import logging as _logging
_logging.getLogger("TitanQ").addHandler(_logging.NullHandler())

__version__ = 'v0.29.0'


# Copyright (c) 2024, InfinityQ Technology, Inc.

# bipolar tools
from ._bipolar_conversion import BipolarToBinary

# mps tools
from ._mps.model_from_mps import from_mps, configure_model_from_mps_file



# Copyright (c) 2024, InfinityQ Technology, Inc.
import logging
import os
import requests

from pydantic import BaseModel, ValidationError
from typing import Any, Dict, Optional, Type
from urllib.parse import urljoin

import requests.adapters
import urllib3

from .model import CreditsResponse, SolveRequest, SolveResponse, TempStorageResponse
from titanq import errors


log = logging.getLogger("TitanQ")

_QUEUED_STATUS = "Queued"
_TITANQ_API_VERSION = "v1"
_USER_AGENT_HEADER = 'User-Agent'

class Client:
    """
    TitanQ api client is a simple wrapper around TitanQ api to help interact with the
    service without the need to deal with http request
    """
    def __init__(self, api_key: Optional[str], base_server_url: str) -> None:
        api_key = api_key or os.getenv("TITANQ_API_KEY")
        if api_key is None:
            raise errors.MissingTitanqApiKey(
                "No API key is provided. You can set your API key in the Model, "
                + "or you can set the environment variable TITANQ_API_KEY")
        self._server_url = base_server_url
        self._api_key = api_key


    def temp_storage(self) -> TempStorageResponse:
        """
        Query temporary storage url's

        :return: The temporary storage response

        :raises requests.exceptions.HTTPError: If an unexpected Error occur during request.
        """
        return self._do_http_request(f"{_TITANQ_API_VERSION}/temp_storage", response_type=TempStorageResponse)

    def credits(self) -> CreditsResponse:
        """
        Query Amount of credits remaining

        :return: The credit response.

        :raises requests.exceptions.HTTPError: If an unexpected Error occur during request.
        """
        return self._do_http_request(f"{_TITANQ_API_VERSION}/credits", response_type=CreditsResponse)

    def solve(self, request: SolveRequest) -> SolveResponse:
        """
        Issue a new solve request to the backend

        :param request: The solve request to issue to the solver

        :return: The response to the solve request (Not the response of the computation)
        """
        log.debug(f"Issuing solve request to TitanQ server ({self._server_url}): {request}")
        response = self._do_http_request(f"{_TITANQ_API_VERSION}/solve", body=request, method='POST', response_type=SolveResponse)

        log.debug(f"Solve request response: {response}")
        return response


    def _do_http_request(
            self,
            path: str,
            *,
            headers: Dict[str, str] = {},
            body: BaseModel = None,
            method='GET',
            response_type: Type[BaseModel]
        ) -> Any:
        """
        Execute the actual http request to the TitanQ api while adding all defaults params

        :param headers: non-default header to the request.
        :param body: Body of the request.
        :param method: Which http method to use while performing the request.
        :param response_type: The object class that the json response will be cast to.

        :raise errors.TitanqError | NotImplementedError: If the request cannot be fulfilled in any way

        :return: The response object created from the json response of the http request.
        """
        headers['authorization'] = self._api_key
        headers[_USER_AGENT_HEADER] = self._user_agent_string()
        url = urljoin(self._server_url, path)
        with requests.Session() as session:
            retries = urllib3.Retry(
                        total=3,
                        backoff_factor=0.5,
                        status_forcelist=[502, 503, 504, 495],
                        allowed_methods={"POST", "GET"},
                    )
            session.mount('https://', requests.adapters.HTTPAdapter(max_retries=retries))
            method = method.upper()

            try:
                if method=='GET':
                    response = session.get(url, headers=headers)
                elif method=='POST':
                    response = session.post(url, headers=headers, data=body.model_dump_json())
                else:
                    raise NotImplementedError(f"Tried to call the TitanQ server with an unknown http method: {method}")
            except requests.exceptions.MissingSchema:
                raise errors.InvalidUrl(f"Invalid TitanQ server URL '{self._server_url}'. Did you mean https://{self._server_url}?")
            except (requests.exceptions.InvalidURL, requests.exceptions.ConnectionError):
                raise errors.InvalidUrl(f"Cannot find a TitanQ server at this URL '{self._server_url}'")
            except Exception as ex:
                raise errors.BadRequest(f"Failed to send a request to the TitanQ server: {ex}")

            self._raise_for_status(response, response_type)

            return self._parse_json(response_type, response.content)


    def _user_agent_string(self) -> str:
        from titanq import __version__ as titanq_version # importing current module without cycle

        request_user_agent = requests.utils.default_headers().get(_USER_AGENT_HEADER, '')
        titanq_user_agent = f"TitanQ-sdk/{titanq_version} " + request_user_agent
        return titanq_user_agent.rstrip()

    def _parse_json(self, json_model, json_string):
        """Create the response object from the response body

        :param json_model: Pydantic model
        :param json_string: JSON string to be parsed as a model instance
        :raises errors.UnexpectedServerResponseError: If JSON string doesn't match model
        :return: response object from the body
        """
        try:
            return json_model.model_validate_json(json_string)
        except ValidationError:
            raise errors.UnexpectedServerResponseError(f"Failed to parse the response from the TitanQ server. The raw response is: {json_string}")

    def _raise_for_status(self, response: requests.Response, response_type):
        """ Raise the right error type depending on the status code of the response passed

        :raises errors.TitanqError: According to the status code
        """
        reason = ""
        try:
            res = self._parse_json(response_type, response.content)
            reason = res.message
        except (AttributeError, errors.UnexpectedServerResponseError):
            # We don't have a solver error message. The HTTP error message will be shown if an error is raised.
            pass

        if response.reason and not reason:
            reason = response.reason

        if 100 <= response.status_code < 400:
            return
        elif 400 == response.status_code:
            raise errors.BadRequest(reason)
        elif 402 == response.status_code:
            raise errors.NotEnoughCreditsError()

        elif 400 <= response.status_code < 500:
            raise errors.ClientError(reason)

        elif 501 == response.status_code:
            raise errors.UnsolvableRequestError()

        elif 500 <= response.status_code < 600:
            raise errors.ServerError(reason)
        else:
            raise errors.UnexpectedServerResponseError(f'{response.status_code} {reason}')



# Copyright (c) 2024, InfinityQ Technology, Inc.
from enum import Enum
import io
import numpy as np

from abc import ABC, abstractmethod
from typing import Optional, Union

from .._client.model import S3Input, S3Output, UrlInput, UrlOutput

class Ftype(Enum):
    WEIGHTS= "weights.npy"
    BIAS = "bias.npy"
    VARIABLE_BOUNDS= "variable_bounds.npy"
    CONSTRAINT_BOUNDS = "constraint_bounds.npy"
    CONSTRAINT_WEIGHTS = "constraints_weights.npy"
    QUAD_CONSTRAINT_WEIGHTS = "quad_constraint_weights.npy"
    QUAD_CONSTRAINT_BOUNDS = "quad_constraint_bounds.npy"
    QUAD_CONSTRAINT_LINEAR_WEIGHTS = "quad_constraint_linear_weights.npy"
    RESULT = "result.zip"


class StorageClient(ABC):
    def __init__(self):
        pass

    @abstractmethod
    def __enter__(self):
        return self

    @abstractmethod
    def __exit__(self, exc_type, exc_value, exc_traceback):
        pass

    @abstractmethod
    def input(self) -> Union[S3Input, UrlInput]:
        """
        Returns the api model for the input of the solve request

        :return: either the s3 or the url input
        """

    @abstractmethod
    def output(self) -> Union[S3Output, UrlOutput]:
        """
        Returns the api model for the output of the solve request

        :return: either the s3 or the url output
        """

    @abstractmethod
    def upload(self, file_type: Ftype, data: bytes):
        """Uploads .npy arrays to the storage client"""

    @abstractmethod
    def wait_for_result_to_be_uploaded_and_download(self) -> bytes:
        """
        Wait until a file is uploaded on the storage client and download it

        :return: content of the result file in bytes
        """

def to_bytes(array: Optional[np.ndarray]) -> Optional[bytes]:
        """
        :return: numpy array as bytes
        """
        if array is None:
            return None

        buffer = io.BytesIO()
        np.save(buffer, array)
        return buffer.getvalue()